@@ -1,5 +1,5 @@
-#[cfg(any(feature = "lua53", feature = "lua51"))]
 #[test]
+#[ignore]
 fn test_compile_fail() {
     let t = trybuild::TestCases::new();
     t.compile_fail("tests/compile_fail/*.rs");
