@@ -1,15 +1,11 @@
-#![cfg_attr(
-    all(feature = "luajit", target_os = "macos", target_arch = "x86_64"),
-    feature(link_args)
-)]
-
-#[cfg_attr(
-    all(feature = "luajit", target_os = "macos", target_arch = "x86_64"),
-    link_args = "-pagezero_size 10000 -image_base 100000000"
-)]
-extern "system" {}
+#![allow(unused_imports)]
 
 use std::panic::catch_unwind;
+use std::rc::Rc;
+use std::time::Duration;
+
+use futures_executor::block_on;
+use futures_util::stream::TryStreamExt;
 
 use mlua::{Error, Function, Lua, Result, Thread, ThreadStatus};
 
@@ -20,13 +16,13 @@ fn test_thread() -> Result<()> {
     let thread = lua.create_thread(
         lua.load(
             r#"
-                function (s)
-                    local sum = s
-                    for i = 1,4 do
-                        sum = sum + coroutine.yield(sum)
-                    end
-                    return sum
+            function (s)
+                local sum = s
+                for i = 1,4 do
+                    sum = sum + coroutine.yield(sum)
                 end
+                return sum
+            end
             "#,
         )
         .eval()?,
@@ -47,11 +43,11 @@ fn test_thread() -> Result<()> {
     let accumulate = lua.create_thread(
         lua.load(
             r#"
-                function (sum)
-                    while true do
-                        sum = sum + coroutine.yield(sum)
-                    end
+            function (sum)
+                while true do
+                    sum = sum + coroutine.yield(sum)
                 end
+            end
             "#,
         )
         .eval::<Function>()?,
@@ -104,6 +100,38 @@ fn test_thread() -> Result<()> {
     Ok(())
 }
 
+#[cfg(feature = "async")]
+#[tokio::test]
+async fn test_thread_stream() -> Result<()> {
+    let lua = Lua::new();
+
+    let thread = lua.create_thread(
+        lua.load(
+            r#"
+            function (s)
+                local sum = s
+                for i = 1,10 do
+                    sum = sum + i
+                    coroutine.yield(sum)
+                end
+                return sum
+            end
+            "#,
+        )
+        .eval()?,
+    )?;
+
+    let mut s = thread.into_async::<_, i64>(0);
+    let mut sum = 0;
+    while let Some(n) = s.try_next().await? {
+        sum += n;
+    }
+
+    assert_eq!(sum, 275);
+
+    Ok(())
+}
+
 #[test]
 fn coroutine_from_closure() -> Result<()> {
     let lua = Lua::new();
@@ -139,3 +167,30 @@ fn coroutine_panic() {
         Err(p) => assert!(*p.downcast::<&str>().unwrap() == "test_panic"),
     }
 }
+
+#[cfg(feature = "async")]
+#[tokio::test]
+async fn test_thread_async() -> Result<()> {
+    let lua = Lua::new();
+
+    let cnt = Rc::new(1); // sleep 1 second
+    let cnt2 = cnt.clone();
+    let f = lua.create_async_function(move |_lua, ()| {
+        let cnt3 = cnt2.clone();
+        async move {
+            futures_timer::Delay::new(Duration::from_secs(*cnt3.as_ref())).await;
+            Ok("hello")
+        }
+    })?;
+
+    let mut thread_s = lua.create_thread(f)?.into_async(());
+    let val: String = thread_s.try_next().await?.unwrap_or_default();
+
+    // thread_s is non-resumable and subject to garbage collection
+
+    lua.gc_collect()?;
+    assert_eq!(Rc::strong_count(&cnt), 1);
+    assert_eq!(val, "hello");
+
+    Ok(())
+}
